# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinMessageSupport library

    @copyright: 2012, 2013 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin.Page import Page
from MoinMoin.log import getLogger
from MoinMoin.user import User
from MoinMoin import wikiutil
from MoinSupport import getHeader, getMetadata, getWikiDict, writeHeaders, \
                        parseDictEntry
from ItemSupport import ItemStore
from MoinMessage import GPG, Message, MoinMessageError, \
                        MoinMessageMissingPart, MoinMessageBadContent, \
                        is_signed, is_encrypted, getContentAndSignature
from email.parser import Parser
import time

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

Dependencies = ['pages']

class MoinMessageAction:

    "Common message handling support for actions."

    def __init__(self, pagename, request):

        """
        On the page with the given 'pagename', use the given 'request' when
        reading posted messages, modifying the Wiki.
        """

        self.pagename = pagename
        self.request = request
        self.page = Page(request, pagename)
        self.init_store()

    def init_store(self):
        self.store = ItemStore(self.page, "messages", "message-locks")

    def do_action(self):
        request = self.request
        content_length = getHeader(request, "Content-Length", "HTTP")
        if content_length:
            content_length = int(content_length)

        self.handle_message_text(request.read(content_length))

    def handle_message_text(self, message_text):

        "Handle the given 'message_text'."

        request = self.request
        message = Parser().parse(StringIO(message_text))

        # Detect any indicated recipient and change the target page, if
        # appropriate.

        if message.has_key("To"):
            try:
                parameters = get_recipient_details(request, message["To"], main=True)
            except MoinMessageRecipientError, exc:
                writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
                request.write("The recipient indicated in the message is not known to this site. "
                              "Details: %s" % exc.message)
                return
            else:
                if parameters["type"] == "page":
                    self.page = Page(request, parameters["location"])
                    self.init_store()

                # NOTE: Support "url".

        # Handle the parsed message.

        self.handle_message(message)

    def handle_message(self, message):

        "Handle the given 'message'."

        # Detect PGP/GPG-encoded payloads.
        # See: http://tools.ietf.org/html/rfc3156

        # Signed payloads are checked and then passed on for further processing
        # elsewhere. Verification is the last step in this base implementation,
        # even if an encrypted-then-signed payload is involved.

        if is_signed(message):
            self.handle_signed_message(message)

        # Encrypted payloads are decrypted and then sent back into this method
        # for signature checking as described above. Thus, signed-then-encrypted
        # payloads are first decrypted and then verified.

        elif is_encrypted(message):
            self.handle_encrypted_message(message)

        # Reject unsigned and unencrypted payloads.

        else:
            request = self.request
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Only PGP/GPG-signed payloads are supported.")

    def handle_encrypted_message(self, message):

        "Handle the given encrypted 'message'."

        request = self.request

        homedir = self.get_homedir()
        if not homedir:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("This site is not configured for this request.")
            return

        gpg = GPG(homedir)

        try:
            text = gpg.decryptMessage(message)

        # Reject messages without a declaration.

        except MoinMessageMissingPart:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("There must be a declaration and a content part for encrypted uploads.")
            return

        # Reject messages without appropriate content.

        except MoinMessageBadContent:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Encrypted data must be provided as application/octet-stream.")
            return

        # Reject any unencryptable message.

        except MoinMessageError:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message could not be decrypted.")
            return

        # Log non-fatal errors.

        if gpg.errors:
            getLogger(__name__).warning(gpg.errors)

        # Handle the embedded message which may itself be a signed message.

        self.handle_message_text(text)

    def handle_signed_message(self, message):

        "Handle the given signed 'message'."

        request = self.request

        # Accept any message whose sender was authenticated by the PGP method.

        if request.user and request.user.valid and request.user.auth_method == "pgp":

            # Handle the embedded message.

            content, signature = getContentAndSignature(message)
            self.handle_message_content(content)

        # Reject any unverified message.

        else:
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message could not be verified. "
                          "Maybe this site is not performing authentication using PGP signatures.")

    def handle_message_content(self, content):

        "Handle the given message 'content'."

        request = self.request

        # Interpret the content as one or more updates.

        message = Message()
        message.handle_message(content)

        # Test any date against the page or message store.

        if message.date:
            store_date = time.gmtime(self.store.mtime())
            page_date = time.gmtime(wikiutil.version2timestamp(self.page.mtime_usecs()))
            last_date = max(store_date, page_date)

            # Reject messages older than the page date.

            if message.date < last_date:
                writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
                request.write("The message is too old: %s versus %s." % (message.date, last_date))
                return

        # Reject messages without dates if so configured.

        elif getattr(request.cfg, "moinmessage_reject_messages_without_dates", True):
            writeHeaders(request, "text/plain", getMetadata(self.page), "403 Forbidden")
            request.write("The message has no date information.")
            return

        # Handle the message as an object.

        self.handle_message_object(message)

    def get_homedir(self):

        "Locate the GPG home directory."

        request = self.request
        homedir = get_homedir(self.request)

        if not homedir:
            writeHeaders(request, "text/plain", getMetadata(self.page), "415 Unsupported Media Type")
            request.write("Encoded data cannot currently be understood. Please notify the site administrator.")

        return homedir

class MoinMessageRecipientError(MoinMessageError):
    pass

class MoinMessageNoRecipients(MoinMessageRecipientError):
    pass

class MoinMessageUnknownRecipient(MoinMessageRecipientError):
    pass

class MoinMessageBadRecipient(MoinMessageRecipientError):
    pass

def get_homedir(request):

    "Locate the GPG home directory."

    return getattr(request.cfg, "moinmessage_gpg_homedir")

def get_signing_users(request):

    "Return a dictionary mapping usernames to signing keys."

    return getWikiDict(
        getattr(request.cfg, "moinmessage_gpg_signing_users_page", "MoinMessageSigningUserDict"),
        request)

def get_relays(request):

    "Return a dictionary mapping relays to URLs."

    return getWikiDict(
        getattr(request.cfg, "moinmessage_gpg_relays_page", "MoinMessageRelayDict"),
        request)

def get_recipients(request, main=False, sending=True, fetching=True):

    """
    Return the recipients dictionary by first obtaining the page in which it
    is stored. This page may either be a subpage of the user's home page, if
    stored on this wiki, or it may be relative to the site root.

    When 'main' is specified and set to a true value, only a dictionary under
    the site root is consulted.

    When 'sending' or 'fetching' is specified and set to a false value, any
    recipients of the indicated type will be excluded from the result of this
    function.

    The name of the subpage is defined by the configuration setting
    'moinmessage_gpg_recipients_page', which if absent is set to
    "MoinMessageRecipientsDict".
    """

    subpage = getattr(request.cfg, "moinmessage_gpg_recipients_page", "MoinMessageRecipientsDict")

    if not main:
        homedetails = wikiutil.getInterwikiHomePage(request)

        if homedetails:
            homewiki, homepage = homedetails
            if homewiki == "Self":
                recipients = getWikiDict("%s/%s" % (homepage, subpage), request)
                if recipients:
                    return filter_recipients(recipients, sending, fetching)

    return filter_recipients(getWikiDict(subpage, request), sending, fetching)

def get_username_for_fingerprint(request, fingerprint):

    """
    Using the 'request', return the username corresponding to the given key
    'fingerprint' or None if no correspondence is present in the mapping page.
    """

    gpg_users = getWikiDict(
        getattr(request.cfg, "moinmessage_gpg_users_page", "MoinMessageUserDict"),
        request,
        superuser=True # disable user test because we have no user yet
        )

    # With a user mapping and a fingerprint corresponding to a known
    # user, temporarily switch user in order to make the edit.

    if gpg_users and gpg_users.has_key(fingerprint):
        return gpg_users[fingerprint]
    else:
        return None

def get_recipient_details(request, recipient, main=False, fetching=False):

    """
    Using the 'request', return a dictionary of details for the specified
    'recipient'. If no details exist, raise a MoinMessageRecipientError
    exception.

    When 'main' is specified and set to a true value, only the recipients
    dictionary under the site root is consulted.

    When 'fetching' is specified and set to a true value, the recipient need
    not have a "type" or "location" defined, but it must have a "fingerprint"
    defined.
    """

    _ = request.getText

    recipients = get_recipients(request, main)
    if not recipients:
        raise MoinMessageNoRecipients, _("No recipients page is defined for MoinMessage.")

    recipient_details = recipients.get(recipient)
    if not recipient_details:
        raise MoinMessageUnknownRecipient, _("The specified recipient is not present in the list of known contacts.")

    parameters = parseDictEntry(recipient_details, ("type", "location", "fingerprint",))

    type = parameters.get("type")
    location = parameters.get("location")
    fingerprint = parameters.get("fingerprint")

    if type in (None, "none") and not fetching:
        raise MoinMessageBadRecipient, _("The recipient details are missing a destination type.")

    if type not in (None, "none") and not location:
        raise MoinMessageBadRecipient, _("The recipient details are missing a location for sent messages.")

    if type in ("url", "relay", None, "none") and not fingerprint:
        raise MoinMessageBadRecipient, _("The recipient details are missing a fingerprint for sending messages.")

    return parameters

def filter_recipients(recipients, sending, fetching):

    """
    Return a copy of the given 'recipients' dictionary retaining all entries
    that apply to the given 'sending' and 'fetching' criteria.
    """

    result = {}
    for recipient, parameters in recipients.items():
        if not fetching and parameters.get("type") in (None, "none"):
            continue
        if not sending and not parameters.get("fingerprint"):
            continue
        result[recipient] = parameters
    return result

# vim: tabstop=4 expandtab shiftwidth=4
